"""
Tests brewblox_ctl.utils
"""

from subprocess import CalledProcessError
from unittest.mock import call

import pytest

from brewblox_ctl import utils

TESTED = utils.__name__


@pytest.fixture
def mocked_ext(mocker):
    mocked = [
        'input',
        'getcwd',
        'getenv_',
        'path',
        'which',
        'machine',
        'check_output',
        'check_call',
    ]
    return {k: mocker.patch(TESTED + '.' + k) for k in mocked}


def test_confirm(mocked_ext):
    mocked_ext['input'].side_effect = [
        '',
        'flapjacks',
        'NNNNO',
        'YES',
    ]
    assert utils.confirm('what?')  # default empty
    assert utils.confirm('why?')  # yes
    assert mocked_ext['input'].call_count == 4


def test_select(mocked_ext):
    m = mocked_ext['input']
    m.return_value = 'answer'
    assert utils.select('question?') == 'answer'
    assert m.call_count == 1

    m.clear_mock()
    m.return_value = ''
    assert utils.select('other question?', 'three') == 'three'
    assert m.call_count == 2


def test_check_ok(mocked_ext):
    m = mocked_ext['check_output']
    assert utils.check_ok('whatever')
    m.side_effect = CalledProcessError(1, '')
    assert not utils.check_ok('really?')

    m.side_effect = RuntimeError()
    with pytest.raises(RuntimeError):
        utils.check_ok('truly?')


def test_path_exists(mocked_ext):
    mocked_ext['path'].exists.side_effect = [
        True,
        False,
    ]
    assert utils.path_exists('p1')
    assert not utils.path_exists('p2')


def test_command_exists(mocked_ext):
    m = mocked_ext['which']
    m.return_value = ''
    assert not utils.command_exists('wobberjacky')

    m.return_value = '/bin/promised-land'
    assert utils.command_exists('pizza')


@pytest.mark.parametrize('combo', [
    ('armv7hf', True),
    ('armv6hf', True),
    ('amd64', False),
    ('x86-64', False)
])
def test_is_pi(combo, mocked_ext):
    machine, result = combo
    mocked_ext['machine'].return_value = machine
    assert utils.is_pi() is result


@pytest.mark.parametrize('combo', [
    ('armv7hf', False),
    ('armv6hf', True),
    ('amd64', False),
    ('x86-64', False)
])
def test_is_v6(combo, mocked_ext):
    machine, result = combo
    mocked_ext['machine'].return_value = machine
    assert utils.is_v6() is result


def test_is_root(mocked_ext):
    assert utils.is_root()

    mocked_ext['check_output'].side_effect = CalledProcessError(1, 'permission denied')
    assert not utils.is_root()


def test_is_docker_user(mocked_ext):
    assert utils.is_docker_user()

    mocked_ext['check_output'].side_effect = CalledProcessError(1, 'not found')
    assert not utils.is_docker_user()


def test_is_brewblox_cwd(mocked_ext):
    mocked_ext['getenv_'].side_effect = [
        '',
        '1.2.3',
    ]
    assert not utils.is_brewblox_cwd()
    assert utils.is_brewblox_cwd()


@pytest.mark.parametrize('combo', [
    ('True', True),
    ('False', False),
    ('false', False),
    ('1', True),
    ('0', False)
])
def test_skipping_confirm(combo, mocked_ext):
    mocked_ext['getenv_'].return_value = combo[0]
    assert utils.skipping_confirm() == combo[1]


def test_optsudo(mocker):
    m = mocker.patch(TESTED + '.is_docker_user')
    m.side_effect = [
        True,
        False,
    ]
    assert utils.optsudo() == ''
    assert utils.optsudo() == 'sudo '


@pytest.mark.parametrize('combo', [
    ('armv6', 'rpi-'),
    ('armv7hf', 'rpi-'),
    ('amd64', ''),
    ('deep-thought', ''),
])
def test_tag_prefix(combo, mocked_ext):
    machine, result = combo
    mocked_ext['machine'].return_value = machine
    assert utils.tag_prefix() == result


def test_docker_tag(mocker):
    mocker.patch(TESTED + '.tag_prefix').return_value = 'prefix'
    mocker.patch(TESTED + '.getenv').return_value = 'value'

    assert utils.docker_tag() == 'prefixvalue'
    assert utils.docker_tag('release') == 'prefixrelease'


@pytest.mark.parametrize('combo', [
    (['armv7hf'], ['', 'edge'], 'rpi-edge'),
    (['amd64'], ['stable'], 'stable'),
    (['deep-thought'], ['stable'], 'stable'),
])
def test_ctl_lib_tag(combo, mocked_ext):
    machine, env, result = combo

    mocked_ext['machine'].side_effect = machine
    mocked_ext['getenv_'].side_effect = env
    assert utils.ctl_lib_tag() == result


def test_check_config(mocked_ext):
    mocked_ext['getenv_'].side_effect = [
        '1.2.3',
        '',
        '',
        '',
        '',
    ]
    mocked_ext['input'].side_effect = [
        '',
        'no',
    ]
    # version ok
    assert utils.check_config()
    # version nok, user ok
    assert not utils.check_config(required=False)
    # required version nok
    with pytest.raises(SystemExit):
        utils.check_config()
    # version nok, user nok
    with pytest.raises(SystemExit):
        utils.check_config(required=False)


def test_prompt_usb(mocked_ext):
    utils.prompt_usb()
    assert mocked_ext['input'].call_count == 1


def test_announce(mocked_ext):
    utils.announce(['cmd1', 'cmd2'])
    assert mocked_ext['input'].call_count == 1


def test_run(mocked_ext):
    utils.run('cmd')
    mocked_ext['check_call'].assert_called_once_with('cmd', shell=True, stderr=utils.STDOUT)


def test_run_all(mocked_ext, mocker):
    s_m = mocker.patch(TESTED + '.skipping_confirm')
    a_m = mocker.patch(TESTED + '.announce')
    r_m = mocker.patch(TESTED + '.run')

    s_m.side_effect = [
        True,
        False,
    ]

    utils.run_all(['cmd1', 'cmd2'], False)
    utils.run_all(['cmd3', 'cmd4'])
    utils.run_all(['cmd5', 'cmd6'])

    assert s_m.call_count == 2

    assert r_m.call_args_list == [
        call('cmd1'),
        call('cmd2'),
        call('cmd3'),
        call('cmd4'),
        call('cmd5'),
        call('cmd6'),
    ]

    assert a_m.call_args_list == [
        call(['cmd5', 'cmd6'])
    ]


def test_lib_loading_command(mocker):
    ctl_lib_tag = mocker.patch(TESTED + '.ctl_lib_tag')
    optsudo = mocker.patch(TESTED + '.optsudo')

    ctl_lib_tag.side_effect = [
        'edge',
        'rpi-stable'
    ]

    optsudo.side_effect = [
        '',
        'sudo ',
    ]

    cmds = utils.lib_loading_commands()
    assert len(cmds) == 6  # no chown
    assert 'sudo' not in ' '.join(cmds)

    cmds = utils.lib_loading_commands()
    assert len(cmds) == 7  # sudo requires a chown
    assert 'sudo' in ' '.join(cmds)



"""
Utility functions
"""

from distutils.util import strtobool
from os import getcwd
from os import getenv as getenv_
from os import path
from platform import machine
from shutil import which
from subprocess import STDOUT, CalledProcessError, check_call, check_output

from brewblox_ctl.const import (CFG_VERSION_KEY, LIB_RELEASE_KEY, RELEASE_KEY,
                                SKIP_CONFIRM_KEY)


def confirm(question, default=True):
    default_val = 'Yes' if default else 'No'
    prompt = '{} [Press ENTER for default value \'{}\']'.format('{}', default_val)
    print(prompt.format(question))
    while True:
        try:
            return bool(strtobool(input().lower() or str(default)))
        except ValueError:
            print('Please type \'y(es)\' or \'n(o)\' and press ENTER.')


def select(question, default=''):
    answer = input('{} {}'.format(
        question,
        '[press ENTER for default value \'{}\']'.format(default) if default else ''))
    return answer or default


def check_ok(cmd):
    try:
        check_output(cmd, shell=True, stderr=STDOUT)
        return True
    except CalledProcessError:
        return False


def getenv(env, default=None):
    return getenv_(env, default)


def path_exists(path_name):
    return path.exists(path_name)


def command_exists(cmd):
    return bool(which(cmd))


def is_pi():
    return machine().startswith('arm')


def is_v6():
    return machine().startswith('armv6')


def is_root():
    return check_ok('ls /root')


def is_docker_user():
    return check_ok('id -nG $USER | grep -qw "docker"')


def is_brewblox_cwd():
    return bool(getenv(CFG_VERSION_KEY))


def skipping_confirm():
    return bool(strtobool(getenv(SKIP_CONFIRM_KEY, 'false')))


def optsudo():
    return 'sudo ' if not is_docker_user() else ''


def tag_prefix():
    return 'rpi-' if is_pi() else ''


def docker_tag(release=None):
    return '{}{}'.format(
        tag_prefix(),
        release or getenv(RELEASE_KEY, 'stable')
    )


def ctl_lib_tag():
    release = getenv(LIB_RELEASE_KEY) or getenv(RELEASE_KEY, 'stable')
    return '{}{}'.format(tag_prefix(), release)


def check_config(required=True):
    if is_brewblox_cwd():
        return True
    elif required:
        print('Please run brewblox-ctl in the same directory as your docker-compose.yml file.')
        raise SystemExit(1)
    elif confirm(
        'No BrewBlox configuration found in current directory ({}). '.format(getcwd()) +
            'Are you sure you want to continue?'):
        return False
    else:
        raise SystemExit(0)


def prompt_usb():
    input('Please press ENTER when your Spark is connected over USB')


def announce(shell_cmds):
    print('The following shell commands will be used: \n')
    for cmd in shell_cmds:
        print('\t', cmd)
    print('')
    input('Press ENTER to continue, Ctrl+C to cancel')


def run(shell_cmd):
    print('\n' + 'Running command: \n\t', shell_cmd, '\n')
    return check_call(shell_cmd, shell=True, stderr=STDOUT)


def run_all(shell_cmds, prompt=True):
    if prompt and not skipping_confirm():
        announce(shell_cmds)
    return [run(cmd) for cmd in shell_cmds]


def lib_loading_commands():
    tag = ctl_lib_tag()
    sudo = optsudo()
    shell_commands = [
        '{}docker rm ctl-lib 2> /dev/null || true'.format(sudo),
        '{}docker pull brewblox/brewblox-ctl-lib:{} || true'.format(sudo, tag),
        '{}docker create --name ctl-lib brewblox/brewblox-ctl-lib:{}'.format(sudo, tag),
        'rm -rf ./brewblox_ctl_lib 2> /dev/null || true ',
        '{}docker cp ctl-lib:/brewblox_ctl_lib ./'.format(sudo),
        '{}docker rm ctl-lib'.format(sudo),
    ]

    if sudo:
        shell_commands += [
            'sudo chown -R $USER ./brewblox_ctl_lib/',
        ]

    return shell_commands



"""
Entrypoint for the BrewBlox commands menu
"""

import sys
from os import getcwd
from subprocess import CalledProcessError

from dotenv import find_dotenv, load_dotenv

from brewblox_ctl import click_helpers, commands, http, utils

HELPTEXT = """
The BrewBlox management tool.

It can be used to create and control BrewBlox configurations.
When used from a BrewBlox installation directory, it will automatically load additional commands.

If the command you're looking for was not found, please check if your current directory
is a BrewBlox installation directory.

By default, BrewBlox is installed to ~/brewblox.

Example use:

    brewblox-ctl install
"""


def check_lib():
    if utils.is_brewblox_cwd() \
        and not utils.path_exists('./brewblox_ctl_lib/__init__.py') \
            and utils.confirm(
                'brewblox-ctl requires extensions that match your BrewBlox release. ' +
                'Do you want to download them now?'):
        utils.run_all(utils.lib_loading_commands())


def local_commands():  # pragma: no cover
    if not utils.is_brewblox_cwd():
        return []

    try:
        check_lib()
        sys.path.append(getcwd())
        from brewblox_ctl_lib import loader
        return loader.cli_sources()

    except ImportError:
        print('No brewblox-ctl extensions found in current directory')
        return []

    except KeyboardInterrupt:
        raise SystemExit(0)

    except CalledProcessError as ex:
        print('\n' + 'Error:', str(ex))
        raise SystemExit(1)


def main():
    try:
        load_dotenv(find_dotenv(usecwd=True))

        if utils.is_root():
            print('brewblox-ctl should not be run as root.')
            raise SystemExit(1)

        if utils.is_v6() \
            and not utils.confirm(
                'Raspberry Pi models 1 and 0 are not supported. Do you want to continue?', False):
            raise SystemExit(0)

        cli = click_helpers.OrderedCommandCollection(
            help=HELPTEXT,
            sources=[
                commands.cli,
                http.cli,
                *local_commands(),
            ])

        cli(standalone_mode=False)

    except Exception as ex:
        print(str(ex), file=sys.stderr)
        raise SystemExit(1)


if __name__ == '__main__':
    main()



"""
Tests brewblox_ctl.__main__
"""

import pytest

from brewblox_ctl import __main__ as main

TESTED = main.__name__


@pytest.fixture(autouse=True)
def mocked_local_commands(mocker):
    m = mocker.patch(TESTED + '.local_commands')
    m.return_value = []
    return m


@pytest.fixture
def mocked_utils(mocker):
    return mocker.patch(TESTED + '.utils')


def test_check_lib(mocked_utils):
    mocked_utils.is_brewblox_cwd.side_effect = [
        True,  # run 1
        True,  # run 2
        True,  # run 3
        False,  # run 4
    ]
    mocked_utils.path_exists.side_effect = [
        False,  # run 1
        False,  # run 2
        True,  # run 3
    ]
    mocked_utils.confirm.side_effect = [
        True,  # run 1
        False,  # run 2
    ]

    main.check_lib()  # ok
    main.check_lib()  # user nok
    main.check_lib()  # files exist
    main.check_lib()  # not a brewblox cwd

    assert mocked_utils.lib_loading_commands.call_count == 1
    assert mocked_utils.run_all.call_count == 1


def test_main(mocked_utils, mocker):
    mock_cli = mocker.patch(TESTED + '.click_helpers.OrderedCommandCollection')
    mocked_utils.is_root.return_value = False
    mocked_utils.is_v6.return_value = False
    main.main()
    assert mock_cli.return_value.call_count == 1


def test_is_root(mocked_utils):
    mocked_utils.is_root.return_value = True
    mocked_utils.is_v6.return_value = False
    with pytest.raises(SystemExit):
        main.main()


def test_is_v6(mocked_utils, mocker):
    mock_cli = mocker.patch(TESTED + '.click_helpers.OrderedCommandCollection')
    mocked_utils.is_root.return_value = False
    mocked_utils.is_v6.return_value = True

    mocked_utils.confirm.side_effect = [
        False,
        True,
    ]

    with pytest.raises(SystemExit):
        main.main()

    main.main()
    assert mock_cli.return_value.call_count == 1


def test_exception(mocked_utils, mocker):
    mock_cli = mocker.patch(TESTED + '.click_helpers.OrderedCommandCollection')
    mock_cli.return_value.side_effect = RuntimeError
    mocked_utils.is_root.return_value = False
    mocked_utils.is_v6.return_value = False

    with pytest.raises(SystemExit):
        main.main()

    assert mock_cli.return_value.call_count == 1



"""
Brewblox-ctl command definitions
"""

import click

from brewblox_ctl import click_helpers, utils
from brewblox_ctl.const import (CFG_VERSION_KEY, PY, RELEASE_KEY,
                                SKIP_CONFIRM_KEY)


def release_tag(release):
    if not release and not utils.is_brewblox_cwd():
        print('Please run this command in a BrewBlox directory, or use the --release argument')
        raise SystemExit(1)

    return utils.docker_tag(release)


@click.group(cls=click_helpers.OrderedGroup)
def cli():
    """Command collector"""


@cli.command()
def down():
    """Stop running services"""
    utils.check_config()
    shell_commands = [
        '{}docker-compose down --remove-orphans'.format(utils.optsudo()),
    ]
    utils.run_all(shell_commands)


@cli.command()
def up():
    """Start all services if not running"""
    utils.check_config()
    shell_commands = [
        '{}docker-compose up -d --remove-orphans'.format(utils.optsudo()),
    ]
    utils.run_all(shell_commands)


@cli.command()
def restart():
    """(re)start all services"""
    utils.check_config()
    sudo = utils.optsudo()
    shell_commands = [
        '{}docker-compose down --remove-orphans'.format(sudo),
        '{}docker-compose up -d'.format(sudo),
    ]
    utils.run_all(shell_commands)


@cli.command()
def install():
    """Install a new BrewBlox system"""
    shell_commands = []

    use_defaults = utils.confirm('Do you want to install with default settings?')

    if utils.command_exists('apt') \
            and (use_defaults or utils.confirm('Do you want to install apt dependency packages?')):
        shell_commands += [
            'sudo apt update',
            'sudo apt upgrade -y',
            'sudo apt install -y libssl-dev libffi-dev',
        ]

    if utils.command_exists('docker'):
        print('Docker is already installed, skipping...')
    elif use_defaults or utils.confirm('Do you want to install Docker?'):
        shell_commands += [
            "curl -sL get.docker.com | sh",
        ]

    if utils.is_docker_user():
        print('{} already belongs to the Docker group, skipping...'.format(utils.getenv('USER')))
    elif use_defaults or utils.confirm('Do you want to run Docker commands without sudo?'):
        shell_commands += [
            'sudo usermod -aG docker $USER'
        ]

    if utils.command_exists('docker-compose'):
        print('docker-compose is already installed, skipping...')
    elif use_defaults or utils.confirm('Do you want to install docker-compose (from pip)?'):
        shell_commands += [
            'sudo {} -m pip install -U docker-compose'.format(PY)
        ]

    if use_defaults:
        target_dir = './brewblox'
    else:
        target_dir = utils.select(
            'In which directory do you want to install the BrewBlox configuration?',
            './brewblox'
        ).rstrip('/')

    if utils.path_exists(target_dir):
        if not utils.confirm('{} already exists. Do you want to continue?'.format(target_dir)):
            return

    # TODO(Bob) Wait until stable is actually stable before offering new users a choice
    release = 'edge'
    # if utils.confirm('Do you want to wait for stable releases?'):
    #     release = 'stable'
    # else:
    #     release = 'edge'

    set_env_command = '{} -m dotenv.cli --quote never -f {}/.env set'.format(PY, target_dir)

    shell_commands += [
        'mkdir -p {}'.format(target_dir),
        'touch {}/.env'.format(target_dir),
        '{} {} {}'.format(set_env_command, RELEASE_KEY, release),
        '{} {} 0.0.0'.format(set_env_command, CFG_VERSION_KEY),
        '{} {} {}'.format(set_env_command, SKIP_CONFIRM_KEY, str(use_defaults)),
    ]

    if use_defaults or utils.confirm('A reboot is recommended. Do you want to do so?'):
        shell_commands.append('sudo reboot')

    utils.run_all(shell_commands, not use_defaults)


@cli.command()
def kill():
    """Stop and remove all containers on this machine"""
    if not utils.confirm('This will stop and remove ALL docker containers on your system. ' +
                         'This includes those not from BrewBlox. ' +
                         'Do you want to continue?'):
        return

    sudo = utils.optsudo()
    shell_commands = [
        '{}docker rm --force $({}docker ps -aq) 2> /dev/null '.format(sudo, sudo) +
        '|| echo "No containers found"',
    ]

    utils.run_all(shell_commands)


@cli.command()
@click.option('--release', default=None, help='BrewBlox release track')
@click.option('--pull/--no-pull', default=True)
def flash(release, pull):
    """Flash firmware on Spark"""
    tag = release_tag(release)
    sudo = utils.optsudo()
    shell_commands = []

    if utils.path_exists('./docker-compose.yml'):
        shell_commands += [
            '{}docker-compose down'.format(sudo),
        ]

    if pull:
        shell_commands += [
            '{}docker pull brewblox/firmware-flasher:{}'.format(sudo, tag),
        ]

    shell_commands += [
        '{}docker run -it --rm --privileged brewblox/firmware-flasher:{} trigger-dfu'.format(sudo, tag),
        '{}docker run -it --rm --privileged brewblox/firmware-flasher:{} flash'.format(sudo, tag),
    ]

    utils.prompt_usb()
    utils.run_all(shell_commands)


@cli.command()
@click.option('--release', default=None, help='BrewBlox release track')
def bootloader(release):
    """Flash bootloader on Spark"""
    tag = release_tag(release)
    sudo = utils.optsudo()
    shell_commands = []

    if utils.path_exists('./docker-compose.yml'):
        shell_commands += [
            '{}docker-compose down'.format(sudo),
        ]

    shell_commands += [
        '{}docker pull brewblox/firmware-flasher:{}'.format(sudo, tag),
        '{}docker run -it --rm --privileged brewblox/firmware-flasher:{} flash-bootloader'.format(
            sudo, tag),
    ]

    utils.prompt_usb()
    utils.run_all(shell_commands)


@cli.command()
@click.option('--release', default=None, help='BrewBlox release track')
def wifi(release):
    """Connect Spark to Wifi"""
    tag = release_tag(release)
    sudo = utils.optsudo()
    shell_commands = []

    if utils.path_exists('./docker-compose.yml'):
        shell_commands += [
            '{}docker-compose down'.format(sudo),
        ]

    shell_commands += [
        '{}docker pull brewblox/firmware-flasher:{}'.format(sudo, tag),
        '{}docker run -it --rm --privileged brewblox/firmware-flasher:{} wifi'.format(sudo, tag),
    ]

    utils.prompt_usb()
    utils.run_all(shell_commands)


@cli.command()
def settings():
    """brewblox-ctl settings"""
    utils.check_config()
    current_skip_setting = utils.skipping_confirm()
    new_skip_setting = utils.confirm(
        'Do you want to skip confirmation prompts when running commands? (is {})'.format(current_skip_setting),
        current_skip_setting)

    shell_commands = [
        '{} -m dotenv.cli --quote never -f .env set {} {}'.format(
            PY, SKIP_CONFIRM_KEY, str(new_skip_setting)),
    ]

    utils.run_all(shell_commands, not new_skip_setting)



"""
Tests brewblox_ctl.commands
"""

from unittest.mock import call

import pytest
from click.testing import CliRunner

from brewblox_ctl import commands
from brewblox_ctl.const import CFG_VERSION_KEY, RELEASE_KEY, SKIP_CONFIRM_KEY

TESTED = commands.__name__


@pytest.fixture
def mocked_py(mocker):
    return mocker.patch(TESTED + '.PY', '/py')


@pytest.fixture
def mocked_utils(mocker):
    m = mocker.patch(TESTED + '.utils')
    m.optsudo.return_value = 'SUDO '
    return m


@pytest.fixture
def mocked_release_tag(mocker):
    m = mocker.patch(TESTED + '.release_tag')
    m.side_effect = lambda r: r or 'tag'
    return m


def test_release_tag(mocked_utils):
    mocked_utils.is_brewblox_cwd.return_value = False
    mocked_utils.docker_tag.side_effect = lambda r: r

    assert commands.release_tag('release') == 'release'
    with pytest.raises(SystemExit):
        commands.release_tag(None)


def test_down(mocked_utils):
    runner = CliRunner()
    result = runner.invoke(commands.down)
    assert result.exit_code == 0

    assert mocked_utils.run_all.call_args_list == [
        call([
            'SUDO docker-compose down --remove-orphans'
        ])
    ]


def test_up(mocked_utils):
    runner = CliRunner()
    result = runner.invoke(commands.up)
    assert result.exit_code == 0

    assert mocked_utils.run_all.call_args_list == [
        call([
            'SUDO docker-compose up -d --remove-orphans'
        ])
    ]


def test_restart(mocked_utils):
    runner = CliRunner()
    result = runner.invoke(commands.restart)
    assert result.exit_code == 0

    assert mocked_utils.run_all.call_args_list == [
        call([
            'SUDO docker-compose down --remove-orphans',
            'SUDO docker-compose up -d',
        ])
    ]


def test_install_express(mocked_utils, mocked_py):
    mocked_utils.command_exists.side_effect = [
        True,  # check apt
        True,  # docker
        True,  # docker-compose
    ]
    mocked_utils.confirm.side_effect = [
        True,  # express install
    ]
    mocked_utils.is_docker_user.side_effect = [
        True,  # optsudo check
        True,
    ]
    mocked_utils.getenv.side_effect = [
        'USEY',
    ]
    mocked_utils.path_exists.side_effect = [
        False,  # target dir
    ]

    runner = CliRunner()
    assert runner.invoke(commands.install).exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.run_all.call_args_list == [
        call([
            'sudo apt update',
            'sudo apt upgrade -y',
            'sudo apt install -y libssl-dev libffi-dev',
            'mkdir -p ./brewblox',
            'touch ./brewblox/.env',
            '/py -m dotenv.cli --quote never -f ./brewblox/.env set {} edge'.format(RELEASE_KEY),
            '/py -m dotenv.cli --quote never -f ./brewblox/.env set {} 0.0.0'.format(CFG_VERSION_KEY),
            '/py -m dotenv.cli --quote never -f ./brewblox/.env set {} True'.format(SKIP_CONFIRM_KEY),
            'sudo reboot',
        ], False)
    ]


def test_install_simple(mocked_utils, mocked_py):
    mocked_utils.command_exists.side_effect = [
        True,  # check apt
        True,  # docker
        True,  # docker-compose
    ]
    mocked_utils.confirm.side_effect = [
        False,  # express install
        False,  # apt update
        True,  # reboot
    ]
    mocked_utils.is_docker_user.side_effect = [
        True,  # optsudo check
        True,
    ]
    mocked_utils.getenv.side_effect = [
        'USEY',
    ]
    mocked_utils.select.side_effect = [
        './brewey//',  # target dir
    ]
    mocked_utils.path_exists.side_effect = [
        False,  # target dir
    ]

    runner = CliRunner()
    assert runner.invoke(commands.install).exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.run_all.call_args_list == [
        call([
            'mkdir -p ./brewey',
            'touch ./brewey/.env',
            '/py -m dotenv.cli --quote never -f ./brewey/.env set {} edge'.format(RELEASE_KEY),
            '/py -m dotenv.cli --quote never -f ./brewey/.env set {} 0.0.0'.format(CFG_VERSION_KEY),
            '/py -m dotenv.cli --quote never -f ./brewey/.env set {} False'.format(SKIP_CONFIRM_KEY),
            'sudo reboot',
        ], True)
    ]


def test_install_decline(mocked_utils, mocked_py):
    mocked_utils.command_exists.side_effect = [
        True,  # check apt
        False,  # docker
        False,  # docker-compose
    ]
    mocked_utils.confirm.side_effect = [
        False,  # express install
        False,  # apt update
        False,  # docker
        False,  # docker user
        False,  # docker-compose
        False,  # reboot
    ]
    mocked_utils.is_docker_user.side_effect = [
        False,  # optsudo check
        False,
    ]
    mocked_utils.getenv.side_effect = [
        'USEY',
    ]
    mocked_utils.select.side_effect = [
        './brewey//',  # target dir
    ]
    mocked_utils.path_exists.side_effect = [
        False,  # target dir
    ]

    runner = CliRunner()
    assert runner.invoke(commands.install).exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.run_all.call_args_list == [
        call([
            'mkdir -p ./brewey',
            'touch ./brewey/.env',
            '/py -m dotenv.cli --quote never -f ./brewey/.env set {} edge'.format(RELEASE_KEY),
            '/py -m dotenv.cli --quote never -f ./brewey/.env set {} 0.0.0'.format(CFG_VERSION_KEY),
            '/py -m dotenv.cli --quote never -f ./brewey/.env set {} False'.format(SKIP_CONFIRM_KEY),
        ], True)
    ]


def test_install_cancel(mocked_utils, mocked_py):
    mocked_utils.command_exists.side_effect = [
        True,  # check apt
        True,  # docker
        True,  # docker-compose
    ]
    mocked_utils.confirm.side_effect = [
        False,  # express install
        False,  # apt update
        False,  # dir exists, don't continue
    ]
    mocked_utils.is_docker_user.side_effect = [
        True,  # optsudo check
        True,
    ]
    mocked_utils.getenv.side_effect = [
        'USEY',
    ]
    mocked_utils.select.side_effect = [
        './brewey//',  # target dir
    ]
    mocked_utils.path_exists.side_effect = [
        True,  # target dir
    ]

    runner = CliRunner()
    assert runner.invoke(commands.install).exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.run_all.call_count == 0


def test_install_all(mocked_utils, mocked_py):
    mocked_utils.command_exists.side_effect = [
        True,  # check apt
        False,  # docker
        False,  # docker-compose
    ]
    mocked_utils.confirm.side_effect = [
        False,  # express install
        True,  # apt update
        True,  # docker
        True,  # docker user
        True,  # docker-compose
        True,  # path exists
        True,  # reboot
    ]
    mocked_utils.is_docker_user.side_effect = [
        False,  # optsudo check
        False,
    ]
    mocked_utils.getenv.side_effect = [
        'USEY',
    ]
    mocked_utils.select.side_effect = [
        './brewey//',  # target dir
    ]
    mocked_utils.path_exists.side_effect = [
        True,  # target dir
    ]

    runner = CliRunner()
    assert runner.invoke(commands.install).exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.run_all.call_count == 1
    args = mocked_utils.run_all.call_args_list[0][0][0]

    assert args == [
        'sudo apt update',
        'sudo apt upgrade -y',
        'sudo apt install -y libssl-dev libffi-dev',
        "curl -sL get.docker.com | sh",
        'sudo usermod -aG docker $USER',
        'sudo /py -m pip install -U docker-compose',
        'mkdir -p ./brewey',
        'touch ./brewey/.env',
        '/py -m dotenv.cli --quote never -f ./brewey/.env set {} edge'.format(RELEASE_KEY),
        '/py -m dotenv.cli --quote never -f ./brewey/.env set {} 0.0.0'.format(CFG_VERSION_KEY),
        '/py -m dotenv.cli --quote never -f ./brewey/.env set {} False'.format(SKIP_CONFIRM_KEY),
        'sudo reboot',
    ]


def test_kill(mocked_utils):
    mocked_utils.confirm.side_effect = [
        False,
        True,
    ]

    runner = CliRunner()
    assert runner.invoke(commands.kill).exit_code == 0

    assert mocked_utils.run_all.call_count == 0

    assert runner.invoke(commands.kill).exit_code == 0
    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.run_all.call_count == 1
    args = mocked_utils.run_all.call_args_list[0][0][0]

    assert args == [
        'SUDO docker rm --force $(SUDO docker ps -aq) 2> /dev/null || echo "No containers found"',
    ]


def test_flash(mocked_utils, mocked_release_tag):
    mocked_utils.path_exists.side_effect = [
        True,
        False,
        False,
    ]

    runner = CliRunner()
    assert runner.invoke(commands.flash).exit_code == 0
    assert runner.invoke(commands.flash, '--release=taggart').exit_code == 0
    assert runner.invoke(commands.flash, '--no-pull').exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.prompt_usb.call_count == 3
    assert mocked_utils.run_all.call_count == 3

    args1 = mocked_utils.run_all.call_args_list[0][0][0]
    args2 = mocked_utils.run_all.call_args_list[1][0][0]
    args3 = mocked_utils.run_all.call_args_list[2][0][0]

    assert args1 == [
        'SUDO docker-compose down',
        'SUDO docker pull brewblox/firmware-flasher:tag',
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:tag trigger-dfu',
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:tag flash',
    ]

    assert args2 == [
        'SUDO docker pull brewblox/firmware-flasher:taggart',
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:taggart trigger-dfu',
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:taggart flash',
    ]

    assert args3 == [
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:tag trigger-dfu',
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:tag flash',
    ]


def test_bootloader(mocked_utils):
    mocked_utils.docker_tag.side_effect = [
        'tag',
        'tag',
    ]
    mocked_utils.path_exists.side_effect = [
        True,
        False,
    ]

    runner = CliRunner()
    assert runner.invoke(commands.bootloader).exit_code == 0
    assert runner.invoke(commands.bootloader).exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.prompt_usb.call_count == 2
    assert mocked_utils.run_all.call_count == 2

    args1 = mocked_utils.run_all.call_args_list[0][0][0]
    args2 = mocked_utils.run_all.call_args_list[1][0][0]

    assert args1 == [
        'SUDO docker-compose down',
        'SUDO docker pull brewblox/firmware-flasher:tag',
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:tag flash-bootloader',
    ]

    assert args2 == args1[1:]


def test_wifi(mocked_utils):
    mocked_utils.docker_tag.side_effect = [
        'tag',
        'tag',
    ]
    mocked_utils.path_exists.side_effect = [
        True,
        False,
    ]

    runner = CliRunner()
    assert runner.invoke(commands.wifi).exit_code == 0
    assert runner.invoke(commands.wifi).exit_code == 0

    assert mocked_utils.check_config.call_count == 0
    assert mocked_utils.prompt_usb.call_count == 2
    assert mocked_utils.run_all.call_count == 2

    args1 = mocked_utils.run_all.call_args_list[0][0][0]
    args2 = mocked_utils.run_all.call_args_list[1][0][0]

    assert args1 == [
        'SUDO docker-compose down',
        'SUDO docker pull brewblox/firmware-flasher:tag',
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:tag wifi',
    ]

    assert args2 == args1[1:]


def test_ctl_settings(mocked_utils, mocked_py):
    mocked_utils.confirm.side_effect = [
        True,
        False,
    ]
    runner = CliRunner()
    assert runner.invoke(commands.settings).exit_code == 0
    assert runner.invoke(commands.settings).exit_code == 0

    args1 = mocked_utils.run_all.call_args_list[0][0][0]
    args2 = mocked_utils.run_all.call_args_list[1][0][0]

    assert args1 == [
        '/py -m dotenv.cli --quote never -f .env set {} True'.format(SKIP_CONFIRM_KEY)
    ]

    assert args2 == [
        '/py -m dotenv.cli --quote never -f .env set {} False'.format(SKIP_CONFIRM_KEY)
    ]



"""
Const values
"""
import sys

PY = sys.executable

CFG_VERSION_KEY = 'BREWBLOX_CFG_VERSION'
LIB_RELEASE_KEY = 'BREWBLOX_CTL_LIB_RELEASE'
RELEASE_KEY = 'BREWBLOX_RELEASE'
SKIP_CONFIRM_KEY = 'BREWBLOX_SKIP_CONFIRM'
